"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function defaultPublish(results) {
    if (results.candidateResult !== results.controlResult ||
        (results.candidateError && !results.controlError) ||
        (!results.candidateError && results.controlError)) {
        console.warn(`Experiment ${results.experimentName}: difference found`);
    }
}
const defaultOptions = {
    publish: defaultPublish
};
/**
 * A factory that creates an experiment function.
 *
 * @param name - The name of the experiment, typically for use in publish.
 * @param control - The legacy function you are trying to replace.
 * @param candidate - The new function intended to replace the control.
 * @param [options] - Options for the experiment. You will usually want to specify a publish function.
 * @returns A function that acts like the control while also running the candidate and publishing results.
 */
function experiment({ name, control, candidate, options = defaultOptions }) {
    const publish = options.publish || defaultPublish;
    return (...args) => {
        let controlResult;
        let candidateResult;
        let controlError;
        let candidateError;
        const isEnabled = !options.enabled || options.enabled(...args);
        function publishResults() {
            if (isEnabled) {
                publish({
                    experimentName: name,
                    experimentArguments: args,
                    controlResult,
                    candidateResult,
                    controlError,
                    candidateError
                });
            }
        }
        if (isEnabled) {
            try {
                candidateResult = candidate(...args);
            }
            catch (e) {
                candidateError = e;
            }
        }
        try {
            controlResult = control(...args);
        }
        catch (e) {
            controlError = e;
            publishResults();
            throw e;
        }
        publishResults();
        return controlResult;
    };
}
exports.experiment = experiment;
/**
 * A factory that creates an asynchronous experiment function.
 *
 * @param name - The name of the experiment, typically for use in publish.
 * @param control - The legacy async function you are trying to replace.
 * @param candidate - The new async function intended to replace the control.
 * @param [options] - Options for the experiment. You will usually want to specify a publish function.
 * @returns An async function that acts like the control while also running the candidate and publishing results.
 */
function experimentAsync({ name, control, candidate, options = defaultOptions }) {
    const publish = options.publish || defaultPublish;
    return async (...args) => {
        let controlResult;
        let candidateResult;
        let controlError;
        let candidateError;
        const isEnabled = !options.enabled || options.enabled(...args);
        function publishResults() {
            if (isEnabled) {
                publish({
                    experimentName: name,
                    experimentArguments: args,
                    controlResult,
                    candidateResult,
                    controlError,
                    candidateError
                });
            }
        }
        if (isEnabled) {
            // Run in parallel
            [candidateResult, controlResult] = await Promise.all([
                candidate(...args).catch((e) => {
                    candidateError = e;
                    return undefined;
                }),
                control(...args).catch((e) => {
                    controlError = e;
                    return undefined;
                })
            ]);
        }
        else {
            controlResult = await control(...args).catch((e) => {
                controlError = e;
                return undefined;
            });
        }
        publishResults();
        if (controlError) {
            throw controlError;
        }
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        return controlResult;
    };
}
exports.experimentAsync = experimentAsync;
